@@ -2,7 +2,7 @@
 pattern.
 
 See
-https://raw.githubusercontent.com/EMMC-ASBL/OntoFlow/master/triplestore/README.md
+https://raw.githubusercontent.com/EMMC-ASBL/tripper/master/README.md
 for an introduction.
 
 This module has no dependencies outside the standard library, but the
@@ -11,340 +11,32 @@ triplestore backends may have.
 For developers: The usage of `s`, `p`, and `o` represent the different parts of an
 RDF Triple: subject, predicate, and object.
 """
+# pylint: disable=invalid-name,too-many-public-methods
 from __future__ import annotations  # Support Python 3.7 (PEP 585)
 
-import hashlib
 import inspect
 import re
 import warnings
 from collections.abc import Sequence
-from datetime import datetime
 from importlib import import_module
 from typing import TYPE_CHECKING
 
+from tripper.errors import NamespaceError, TriplestoreError, UniquenessError
+from tripper.literal import Literal
+from tripper.namespace import DCTERMS, DM, FNO, MAP, OWL, RDF, RDFS, XML, XSD, Namespace
+from tripper.utils import en, function_id, infer_iri, split_iri
+
 if TYPE_CHECKING:  # pragma: no cover
     from collections.abc import Mapping
-    from typing import Any, Callable, Dict, Generator, Tuple, Union
+    from typing import Callable, Dict, Generator, List, Optional, Tuple, Union
 
-    Triple = Tuple[Union[str, None], Union[str, None], Union[str, None]]
+    from tripper.utils import OptionalTriple, Triple
 
 
 # Regular expression matching a prefixed IRI
 _MATCH_PREFIXED_IRI = re.compile(r"^([a-z]+):([^/]{2}.*)$")
 
 
-class TriplestoreError(Exception):
-    """Base exception for triplestore errors."""
-
-
-class UniquenessError(TriplestoreError):
-    """More than one matching triple."""
-
-
-class NamespaceError(TriplestoreError):
-    """Namespace error."""
-
-
-class NoSuchIRIError(NamespaceError):
-    """Namespace has no such IRI."""
-
-
-class Namespace:
-    """Represent a namespace.
-
-    Arguments:
-        iri: IRI of namespace to represent.
-        label_annotations: Sequence of label annotations. If given, check
-            the underlying ontology during attribute access if the name
-            correspond to a label. The label annotations should be ordered
-            from highest to lowest precedense.
-            If True is provided, `label_annotations` is set to
-            ``(SKOS.prefLabel, RDF.label, SKOS.altLabel)``.
-        check: Whether to check underlying ontology if the IRI exists during
-            attribute access.  If true, NoSuchIRIError will be raised if the
-            IRI does not exist in this namespace.
-        cachemode: Should be one of:
-              - Namespace.NO_CACHE: Turn off caching.
-              - Namespace.USE_CACHE: Cache attributes as they are looked up.
-              - Namespace.ONLY_CACHE: Cache all names at initialisation time.
-                Do not access the triplestore after that.
-            Default is `NO_CACHE` if neither `label_annotations` or `check`
-            is given, otherwise `USE_CACHE`.
-        triplestore: Use this triplestore for label lookup and checking.
-            If not given, and either `label_annotations` or `check` are
-            enabled, a new rdflib triplestore will be created.
-        triplestore_url: Alternative URL to use for loading the underlying
-            ontology if `triplestore` is not given.  Defaults to `iri`.
-    """
-
-    NO_CACHE = 0
-    USE_CACHE = 1
-    ONLY_CACHE = 2
-
-    __slots__ = (
-        "_iri",
-        "_label_annotations",
-        "_check",
-        "_cache",
-        "_triplestore",
-    )
-
-    def __init__(
-        self,
-        iri,
-        label_annotations=(),
-        check=False,
-        cachemode=-1,
-        triplestore=None,
-        triplestore_url=None,
-    ):
-        if label_annotations is True:
-            label_annotations = (SKOS.prefLabel, RDF.label, SKOS.altLabel)
-
-        self._iri = str(iri)
-        self._label_annotations = tuple(label_annotations)
-        self._check = bool(check)
-
-        need_triplestore = bool(check or label_annotations)
-        if cachemode == -1:
-            cachemode = Namespace.ONLY_CACHE if need_triplestore else Namespace.NO_CACHE
-
-        if need_triplestore and triplestore is None:
-            url = triplestore_url if triplestore_url else iri
-            triplestore = Triplestore("rdflib", base_iri=iri)
-            triplestore.parse(url)
-
-        self._cache = {} if cachemode != Namespace.NO_CACHE else None
-        #
-        # FIXME:
-        # Change this to only assigning the triplestore if cachemode is
-        # ONLY_CACHE when we figure out a good way to pre-populate the
-        # cache with IRIs from the triplestore.
-        #
-        # self._triplestore = (
-        #    triplestore if cachemode != Namespace.ONLY_CACHE else None
-        # )
-        self._triplestore = triplestore if need_triplestore else None
-
-        if cachemode != Namespace.NO_CACHE:
-            self._update_cache(triplestore)
-
-    def _update_cache(self, triplestore=None):
-        """Update the internal cache from `triplestore`."""
-        if not triplestore:
-            triplestore = self._triplestore
-        if not triplestore:
-            raise NamespaceError("`triplestore` argument needed for updating the cache")
-        if self._cache is None:
-            self._cache = {}
-
-        # Add (label, full_iri) pairs to cache
-        for label in reversed(self._label_annotations):
-            self._cache.update(
-                (o, s)
-                for s, o in triplestore.subject_objects(label)
-                if s.startswith(self._iri)
-            )
-
-        # Add (name, full_iri) pairs to cache
-        # Currently we only check concepts that defines RDFS.isDefinedBy
-        # relations.
-        # Is there an efficient way to loop over all IRIs in this namespace?
-        self._cache.update(
-            (s[len(self._iri) :], s)
-            for s in triplestore.subjects(RDFS.isDefinedBy, self._iri)
-            if s.startswith(self._iri)
-        )
-
-    def __getattr__(self, name):
-        if self._cache and name in self._cache:
-            return self._cache[name]
-
-        if self._triplestore:
-
-            # Check if ``iri = self._iri + name`` is in the triplestore.
-            # If so, add it to the cache.
-            # We only need to check that generator returned by
-            # `self._triplestore.predicate_objects(iri)` is non-empty.
-            iri = self._iri + name
-            predicate_object = self._triplestore.predicate_objects(iri)
-            try:
-                predicate_object.__next__()
-            except StopIteration:
-                pass
-            else:
-                if self._cache is not None:
-                    self._cache[name] = iri
-                return iri
-
-            # Check for label annotations matching `name`.
-            for label in self._label_annotations:
-                for s, o in self._triplestore.subject_objects(label):
-                    if o == name and s.startswith(self._iri):
-                        if self._cache is not None:
-                            self._cache[name] = s
-                        return s
-
-        if self._check:
-            raise NoSuchIRIError(self._iri + name)
-        return self._iri + name
-
-    def __getitem__(self, key):
-        return self.__getattr__(key)
-
-    def __repr__(self):
-        return f"Namespace({self._iri})"
-
-    def __str__(self):
-        return self._iri
-
-    def __add__(self, other):
-        return self._iri + str(other)
-
-
-# Pre-defined namespaces
-XML = Namespace("http://www.w3.org/XML/1998/namespace")
-RDF = Namespace("http://www.w3.org/1999/02/22-rdf-syntax-ns#")
-RDFS = Namespace("http://www.w3.org/2000/01/rdf-schema#")
-XSD = Namespace("http://www.w3.org/2001/XMLSchema#")
-OWL = Namespace("http://www.w3.org/2002/07/owl#")
-SKOS = Namespace("http://www.w3.org/2004/02/skos/core#")
-DC = Namespace("http://purl.org/dc/elements/1.1/")
-DCTERMS = Namespace("http://purl.org/dc/terms/")
-FOAF = Namespace("http://xmlns.com/foaf/0.1/")
-DOAP = Namespace("http://usefulinc.com/ns/doap#")
-PROV = Namespace("http://www.w3.org/ns/prov#")
-DCAT = Namespace("http://www.w3.org/ns/dcat#")
-TIME = Namespace("http://www.w3.org/2006/time#")
-FNO = Namespace("https://w3id.org/function/ontology#")
-
-EMMO = Namespace("http://emmo.info/emmo#")
-MAP = Namespace("http://emmo.info/domain-mappings#")
-DM = Namespace("http://emmo.info/datamodel#")
-
-
-class Literal(str):
-    """A literal RDF value."""
-
-    lang: str
-    datatype: "Any"
-
-    def __new__(cls, value, lang=None, datatype=None):
-        string = super().__new__(cls, value)
-        if lang:
-            if datatype:
-                raise TypeError("A literal can only have one of `lang` or `datatype`.")
-            string.lang = str(lang)
-            string.datatype = None
-        else:
-            string.lang = None
-            if datatype:
-                string.datatype = {
-                    str: XSD.string,
-                    bool: XSD.boolean,
-                    int: XSD.integer,
-                    float: XSD.double,
-                    bytes: XSD.hexBinary,
-                    bytearray: XSD.hexBinary,
-                    datetime: XSD.dateTime,
-                }.get(datatype, datatype)
-            elif isinstance(value, str):
-                string.datatype = None
-            elif isinstance(value, bool):
-                string.datatype = XSD.boolean
-            elif isinstance(value, int):
-                string.datatype = XSD.integer
-            elif isinstance(value, float):
-                string.datatype = XSD.double
-            elif isinstance(value, (bytes, bytearray)):
-                string = value.hex()
-                string.datatype = XSD.hexBinary
-            elif isinstance(value, datetime):
-                string.datatype = XSD.dateTime
-                # TODO:
-                #   - XSD.base64Binary
-                #   - XSD.byte, XSD.unsignedByte
-            else:
-                string.datatype = None
-        return string
-
-    def __repr__(self):
-        lang = f", lang='{self.lang}'" if self.lang else ""
-        datatype = f", datatype='{self.datatype}'" if self.datatype else ""
-        return f"Literal('{self}'{lang}{datatype})"
-
-    value = property(
-        fget=lambda self: self.to_python(),
-        doc="Appropriate python datatype derived from this RDF literal.",
-    )
-
-    def to_python(self):
-        """Returns an appropriate python datatype derived from this RDF
-        literal."""
-        value = str(self)
-
-        if self.datatype == XSD.boolean:
-            value = bool(self)
-        elif self.datatype in (
-            XSD.integer,
-            XSD.int,
-            XSD.short,
-            XSD.long,
-            XSD.nonPositiveInteger,
-            XSD.negativeInteger,
-            XSD.nonNegativeInteger,
-            XSD.unsignedInt,
-            XSD.unsignedShort,
-            XSD.unsignedLong,
-            XSD.byte,
-            XSD.unsignedByte,
-        ):
-            value = int(self)
-        elif self.datatype in (
-            XSD.double,
-            XSD.decimal,
-            XSD.dataTimeStamp,
-            OWL.real,
-            OWL.rational,
-        ):
-            value = float(self)
-        elif self.datatype == XSD.hexBinary:
-            value = self.encode()
-        elif self.datatype == XSD.dateTime:
-            value = datetime.fromisoformat(self)
-        elif self.datatype and self.datatype not in (
-            RDF.PlainLiteral,
-            RDF.XMLLiteral,
-            RDFS.Literal,
-            XSD.anyURI,
-            XSD.language,
-            XSD.Name,
-            XSD.NMName,
-            XSD.normalizedString,
-            XSD.string,
-            XSD.token,
-            XSD.NMTOKEN,
-        ):
-            warnings.warn(f"unknown datatype: {self.datatype} - assuming string")
-        return value
-
-    def n3(self):  # pylint: disable=invalid-name
-        """Returns a representation in n3 format."""
-        if self.lang:
-            return f'"{self}"@{self.lang}'
-        if self.datatype:
-            return f'"{self}"^^{self.datatype}'
-        return f'"{self}"'
-
-
-def en(value):  # pylint: disable=invalid-name
-    """Convenience function that returns value as a plain english literal.
-
-    Equivalent to``Literal(value, lang="en")``.
-    """
-    return Literal(value, lang="en")
-
-
 class Triplestore:
     """Provides a common frontend to a range of triplestore backends."""
 
@@ -365,49 +57,78 @@ class Triplestore:
         # "dm": DM,
     }
 
-    def __init__(self, backend: str, base_iri: str = None, **kwargs):
+    def __init__(
+        self,
+        backend: str,
+        base_iri: "Optional[str]" = None,
+        database: "Optional[str]" = None,
+        **kwargs,
+    ) -> None:
         """Initialise triplestore using the backend with the given name.
 
         Parameters:
             backend: Name of the backend module.
             base_iri: Base IRI used by the add_function() method when adding
-                new triples.
+                new triples. May also be used by the backend.
+            database: Name of database to connect to (for backends that supports it).
             kwargs: Keyword arguments passed to the backend's __init__()
                 method.
         """
         module = import_module(
             backend if "." in backend else f"tripper.backends.{backend}"
         )
-        cls = getattr(module, backend.title() + "Strategy")
+        cls = getattr(module, f"{backend.title()}Strategy")
         self.base_iri = base_iri
         self.namespaces: "Dict[str, Namespace]" = {}
+        self.closed = False
         self.backend_name = backend
-        self.backend = cls(base_iri=base_iri, **kwargs)
+        self.backend = cls(base_iri=base_iri, database=database, **kwargs)
         # Keep functions in the triplestore for convienence even though
         # they usually do not belong to the triplestore per se.
-        self.function_repo: "Dict[str, Union[float, Callable[[], float]]]" = {}
+        self.function_repo: "Dict[str, Union[float, Callable[[], float], None]]" = {}
         for prefix, namespace in self.default_namespaces.items():
             self.bind(prefix, namespace)
 
     # Methods implemented by backend
     # ------------------------------
-    def triples(self, triple: "Triple") -> "Generator":
-        """Returns a generator over matching triples."""
+    def triples(self, triple: "Triple") -> "Generator[Triple, None, None]":
+        """Returns a generator over matching triples.
+
+        Arguments:
+            triple: A `(s, p, o)` tuple where `s`, `p` and `o` should
+                either be None (matching anything) or an exact IRI to
+                match.
+        """
         return self.backend.triples(triple)
 
     def add_triples(self, triples: "Sequence[Triple]"):
-        """Add a sequence of triples."""
+        """Add a sequence of triples.
+
+        Arguments:
+            triples: A sequence of `(s, p, o)` tuples to add to the
+                triplestore.
+        """
         self.backend.add_triples(triples)
 
-    def remove(self, triple: "Triple"):
+    def remove(self, triple: "Triple") -> None:
         """Remove all matching triples from the backend."""
         self.backend.remove(triple)
 
     # Methods optionally implemented by backend
     # -----------------------------------------
+    def close(self):
+        """Calls the backend close() method if it is implemented.
+        Otherwise, this method has no effect.
+        """
+        # It should be ok to call close() regardless of whether the backend
+        # implements this method or not.  Hence, don't call _check_method().
+        if not self.closed and hasattr(self.backend, "close"):
+            self.backend.close()
+        self.closed = True
+
     def parse(
         self, source=None, format=None, **kwargs  # pylint: disable=redefined-builtin
-    ):
+    ) -> None:
         """Parse source and add the resulting triples to triplestore.
 
         Parameters:
@@ -447,17 +168,42 @@ class Triplestore:
         self._check_method("serialize")
         return self.backend.serialize(destination=destination, format=format, **kwargs)
 
-    def query(self, query_object, **kwargs):
-        """SPARQL query."""
+    def query(self, query_object, **kwargs) -> "List[Tuple[str, ...]]":
+        """SPARQL query.
+
+        Parameters:
+            query_object: String with the SPARQL query.
+            kwargs: Keyword arguments passed to the backend query() method.
+
+        Returns:
+            List of tuples of IRIs for each matching row.
+
+        Note:
+            This method is intended for SELECT queries. Use
+            the update() method for INSERT and DELETE queries.
+
+        """
         self._check_method("query")
         return self.backend.query(query_object=query_object, **kwargs)
 
-    def update(self, update_object, **kwargs):
-        """Update triplestore with SPARQL."""
+    def update(self, update_object, **kwargs) -> None:
+        """Update triplestore with SPARQL.
+
+        Parameters:
+            update_object: String with the SPARQL query.
+            kwargs: Keyword arguments passed to the backend update() method.
+
+        Note:
+            This method is intended for INSERT and DELETE queries. Use
+            the query() method for SELECT queries.
+
+        """
         self._check_method("update")
         return self.backend.update(update_object=update_object, **kwargs)
 
-    def bind(self, prefix: str, namespace: "Union[str, Namespace]", **kwargs):
+    def bind(
+        self, prefix: str, namespace: "Union[str, Namespace]", **kwargs
+    ) -> Namespace:
         """Bind prefix to namespace and return the new Namespace object.
 
         The new Namespace is created with `namespace` as IRI.
@@ -479,26 +225,97 @@ class Triplestore:
         )
         return self.namespaces[prefix]
 
+    @classmethod
+    def create_database(cls, backend: str, database: str, **kwargs):
+        """Create a new database in backend.
+
+        Parameters:
+            backend: Name of backend.
+            database: Name of the new database.
+            kwargs: Keyword arguments passed to the backend
+                create_database() method.
+
+        Note:
+            This is a class method, which operates on the backend
+            triplestore without connecting to it.
+        """
+        cls._check_backend_method(backend, "create_database")
+        backend_class = cls._get_backend(backend)
+        return backend_class.create_database(database=database, **kwargs)
+
+    @classmethod
+    def remove_database(cls, backend: str, database: str, **kwargs):
+        """Remove a database in backend.
+
+        Parameters:
+            backend: Name of backend.
+            database: Name of the database to be removed.
+            kwargs: Keyword arguments passed to the backend
+                remove_database() method.
+
+        Note:
+            This is a class method, which operates on the backend
+            triplestore without connecting to it.
+        """
+        cls._check_backend_method(backend, "remove_database")
+        backend_class = cls._get_backend(backend)
+        return backend_class.remove_database(database=database, **kwargs)
+
+    @classmethod
+    def list_databases(cls, backend: str, **kwargs):
+        """For backends that supports multiple databases, list of all
+        databases.
+
+        Parameters:
+            backend: Name of backend.
+            kwargs: Keyword arguments passed to the backend
+                list_databases() method.
+
+        Note:
+            This is a class method, which operates on the backend
+            triplestore without connecting to it.
+        """
+        cls._check_backend_method(backend, "list_databases")
+        backend_class = cls._get_backend(backend)
+        return backend_class.list_databases(**kwargs)
+
     # Convenient methods
     # ------------------
     # These methods are modelled after rdflib and provide some convinient
     # interfaces to the triples(), add_triples() and remove() methods
     # implemented by all backends.
+
+    @classmethod
+    def _get_backend(cls, backend: str):
+        """Returns the class implementing the given backend."""
+        module = import_module(
+            backend if "." in backend else f"tripper.backends.{backend}"
+        )
+        return getattr(module, f"{backend.title()}Strategy")
+
+    @classmethod
+    def _check_backend_method(cls, backend: str, name: str):
+        """Checks that `backend` has a method called `name`.
+
+        Raises NotImplementedError if it hasn't.
+        """
+        backend_class = cls._get_backend(backend)
+        if not hasattr(backend_class, name):
+            raise NotImplementedError(
+                f'Triplestore backend {backend!r} do not implement a "{name}()" method.'
+            )
+
     def _check_method(self, name):
         """Check that backend implements the given method."""
-        if not hasattr(self.backend, name):
-            raise NotImplementedError(
-                f"Triplestore backend {self.backend_name!r} do not implement a "
-                f'"{name}()" method.'
-            )
+        self._check_backend_method(self.backend_name, name)
 
     def add(self, triple: "Triple"):
         """Add `triple` to triplestore."""
         self.add_triples([triple])
 
-    def value(
+    def value(  # pylint: disable=redefined-builtin
         self, subject=None, predicate=None, object=None, default=None, any=False
-    ):  # pylint: disable=redefined-builtin
+    ):
         """Return the value for a pair of two criteria.
 
         Useful if one knows that there may only be one value.
@@ -613,12 +430,12 @@ class Triplestore:
                 raise NamespaceError(f"No prefix defined for IRI: {iri}")
         return iri
 
-    def add_mapsTo(  # pylint: disable=invalid-name
+    def add_mapsTo(
         self,
         target: str,
         source: str,
-        property_name: str = None,
-        cost: "Union[float, Callable]" = None,
+        property_name: "Optional[str]" = None,
+        cost: "Optional[Union[float, Callable]]" = None,
         target_cost: bool = True,
     ):
         """Add 'mapsTo' relation to triplestore.
@@ -645,7 +462,8 @@ class Triplestore:
         target = self.expand_iri(target)
         source = self.expand_iri(infer_iri(source))
         if property_name:
-            source = f"{source}#{property_name}"
+            self.add((f"{source}#{property_name}", MAP.mapsTo, target))
+        else:
             self.add((source, MAP.mapsTo, target))
         if cost is not None:
             dest = target if target_cost else source
@@ -653,17 +471,18 @@ class Triplestore:
 
     def add_function(
         self,
-        func: Callable,
+        func: "Union[Callable, str]",
         expects: "Union[str, Sequence, Mapping]" = (),
         returns: "Union[str, Sequence]" = (),
-        base_iri: str = None,
+        base_iri: "Optional[str]" = None,
         standard: str = "fno",
-        cost: "Union[float, Callable]" = None,
+        cost: "Optional[Union[float, Callable]]" = None,
     ):
         """Inspect function and add triples describing it to the triplestore.
 
         Parameters:
-            func: Function to describe.
+            func: Function to describe.  Should either be a callable or a
+                string with a unique function IRI.
             expects: Sequence of IRIs to ontological concepts corresponding
                 to positional arguments of `func`.  May also be given as a
                 dict mapping argument names to corresponding ontological IRIs.
@@ -688,11 +507,9 @@ class Triplestore:
 
         method = getattr(self, f"_add_function_{standard}")
         func_iri = method(func, expects, returns, base_iri)
-        self.function_repo[func_iri] = func
-
+        self.function_repo[func_iri] = func if callable(func) else None
         if cost is not None:
-            for dest_iri in returns:
-                self._add_cost(cost, dest_iri)
+            self._add_cost(cost, func_iri)
 
         return func_iri
 
@@ -724,24 +541,39 @@ class Triplestore:
 
         if base_iri is None:
             base_iri = self.base_iri if self.base_iri else ":"
-        fid = function_id(func)  # Function id
-        doc_string = inspect.getdoc(func)
-        func_iri = f"{base_iri}{func.__name__}_{fid}"
-        parlist = f"_:{func.__name__}{fid}parlist"
-        outlist = f"_:{func.__name__}{fid}outlist"
+
+        if callable(func):
+            fid = function_id(func)  # Function id
+            func_iri = f"{base_iri}{func.__name__}_{fid}"
+            doc_string = inspect.getdoc(func)
+            parlist = f"_:{func.__name__}{fid}_parlist"
+            outlist = f"_:{func.__name__}{fid}_outlist"
+            if isinstance(expects, Sequence):
+                pars = list(zip(expects, inspect.signature(func).parameters))
+            else:
+                pars = [
+                    (expects[par], par) for par in inspect.signature(func).parameters
+                ]
+        elif isinstance(func, str):
+            func_iri = func
+            doc_string = ""
+            parlist = f"_:{func_iri}_parlist"
+            outlist = f"_:{func_iri}_outlist"
+            pariris = expects if isinstance(expects, Sequence) else expects.values()
+            parnames = [split_iri(pariri)[1] for pariri in pariris]
+            pars = list(zip(pariris, parnames))
+        else:
+            raise TypeError("`func` should be either a callable or an IRI")
+
         self.add((func_iri, RDF.type, FNO.Function))
         self.add((func_iri, FNO.expects, parlist))
         self.add((func_iri, FNO.returns, outlist))
         if doc_string:
             self.add((func_iri, DCTERMS.description, en(doc_string)))
 
-        if isinstance(expects, Sequence):
-            items = list(zip(expects, inspect.signature(func).parameters))
-        else:
-            items = [(expects[par], par) for par in inspect.signature(func).parameters]
         lst = parlist
-        for i, (iri, parname) in enumerate(items):
-            lst_next = f"{parlist}{i+2}" if i < len(items) - 1 else RDF.nil
+        for i, (iri, parname) in enumerate(pars):
+            lst_next = f"{parlist}{i+2}" if i < len(pars) - 1 else RDF.nil
             par = f"{func_iri}_parameter{i+1}_{parname}"
             self.add((par, RDF.type, FNO.Parameter))
             self.add((par, RDFS.label, en(parname)))
@@ -761,39 +593,3 @@ class Triplestore:
             lst = lst_next
 
         return func_iri
-
-
-def infer_iri(obj):
-    """Return IRI of the individual that stands for object `obj`."""
-    if isinstance(obj, str):
-        return obj
-    if hasattr(obj, "uri") and obj.uri:
-        # dlite.Metadata or dataclass (or instance with uri)
-        return obj.uri
-    if hasattr(obj, "uuid") and obj.uuid:
-        # dlite.Instance or dataclass
-        return obj.uuid
-    if hasattr(obj, "schema") and callable(obj.schema):
-        # pydantic.BaseModel
-        schema = obj.schema()
-        properties = schema["properties"]
-        if "uri" in properties and properties["uri"]:
-            return properties["uri"]
-        if "uuid" in properties and properties["uuid"]:
-            return properties["uuid"]
-    raise TypeError("cannot infer IRI from object {obj!r}")
-
-
-def function_id(func, length=4):
-    """Return a checksum for function `func`.
-
-    The returned object is a string of hexadecimal digits.
-
-    `length` is the number of bytes in the returned checksum.  Since
-    the current implementation is based on the shake_128 algorithm,
-    it make no sense to set `length` larger than 32 bytes.
-    """
-    # return hex(crc32(inspect.getsource(func).encode())).lstrip('0x')
-    return hashlib.shake_128(  # pylint: disable=too-many-function-args
-        inspect.getsource(func).encode()
-    ).hexdigest(length)
